//
//  Augmented Reality Courses
//   (USING OPENCV 3)
//
//  Created by Thierry CHATEAU on 26/09/2016.
//  Copyright (c) 2016 Thierry CHATEAU. All rights reserved.
//
#include <stdio.h>
#include <iostream>
#include <opencv2/features2d.hpp>
#include <opencv2/xfeatures2d.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/opencv.hpp>
#include <vector>
#include <iostream>

using namespace cv;
using namespace std;
using namespace cv::xfeatures2d;

/** @function main */
int main( int argc, char** argv )
{
    // Read newspaper image
    Mat img_scene = imread( "journal2.png", CV_LOAD_IMAGE_GRAYSCALE );
    // Read object and AR images
    Mat img_object = imread( "im_req.png", CV_LOAD_IMAGE_GRAYSCALE );
    Mat img_warp = imread( "im_ra.png", CV_LOAD_IMAGE_GRAYSCALE );
    // Resize AR Image
    resize(img_warp,img_warp,img_object.size());
    std::vector<KeyPoint> keypoints_object;
    // ==================
    // 1) Define detector and detect object interest points
    // Detector
    Ptr<Feature2D> detector = SIFT::create();
    detector->detect( img_object, keypoints_object ); // compute interest points
    //SurfDescriptorExtractor extractor; // descriptor declaration
    Mat descriptors_object; // object descriptor matrix declaration
    detector->compute( img_object, keypoints_object, descriptors_object ); // Compute object descriptors
    double t = (double)getTickCount();
    //-- Step 1: Detect the keypoints using Detector
    std::vector<KeyPoint> keypoints_scene;
    detector->detect( img_scene, keypoints_scene );
        
    //-- Step 2: Calculate descriptors (feature vectors)
    Mat descriptors_scene;
    detector->compute( img_scene, keypoints_scene, descriptors_scene );
        
    //-- Step 3: Matching descriptor vectors using a matcher
    FlannBasedMatcher matcher;
    std::vector< DMatch > matches;
    matcher.match( descriptors_object, descriptors_scene, matches );
        
    //-- Quick calculation of max and min distances between keypoints
    double max_dist = 0; double min_dist = 100;
    for( int i = 0; i < descriptors_object.rows; i++ )
    { double dist = matches[i].distance;
        if( dist < min_dist ) min_dist = dist;
        if( dist > max_dist ) max_dist = dist;
    }
        
    printf("-- Max dist : %f \n", max_dist );
    printf("-- Min dist : %f \n", min_dist );
        
    //-- Keep only "good" matches (i.e. whose distance is less than n*min_dist )
    unsigned int n=3;
    std::vector< DMatch > good_matches;
    for( int i = 0; i < descriptors_object.rows; i++ )
    { if( matches[i].distance < n*min_dist )
    { good_matches.push_back( matches[i]); }
    }
    // Compute and display processing time
    t = ((double)getTickCount() - t)/getTickFrequency();
    cout << "Times passed in seconds: " << t << endl;
    //-- Localize the object
    std::vector<Point2f> obj;
    std::vector<Point2f> scene;
        
    for( int i = 0; i < good_matches.size(); i++ )
    {
        //-- Get the keypoints from the good matches
        obj.push_back( keypoints_object[ good_matches[i].queryIdx ].pt );
        scene.push_back( keypoints_scene[ good_matches[i].trainIdx ].pt );
    }
    // Estimate an Homography (if at least 5 matches exist)
    if(obj.size()>5){
        // .. To be complete by student    
        Mat H = //...
        // warp
        Mat dst;
        // .. To be complete by student 
        // Mix real image with AR im
        addWeighted(dst,0.5,img_scene,0.5,0.0,dst);
        // Draw matches
        Mat img_matches;
        drawMatches( img_object, keypoints_object, img_scene, keypoints_scene,
                good_matches, img_matches, Scalar::all(-1), Scalar::all(-1),
                vector<char>(), DrawMatchesFlags::NOT_DRAW_SINGLE_POINTS );
        
        //////////////////
        // DISPLAY
        
        //-- Get the corners from the image_1 ( the object to be "detected" )
        std::vector<Point2f> obj_corners(4);
        obj_corners[0] = cvPoint(0,0); obj_corners[1] = cvPoint( img_object.cols, 0 );
        obj_corners[2] = cvPoint( img_object.cols, img_object.rows ); obj_corners[3] = cvPoint( 0, img_object.rows );
        std::vector<Point2f> scene_corners(4);
        
        perspectiveTransform( obj_corners, scene_corners, H);
        
        //-- Draw lines between the corners (the mapped object in the scene - image_2 )
        line( img_matches, scene_corners[0] + Point2f( img_object.cols, 0), scene_corners[1] + Point2f( img_object.cols, 0), Scalar(0, 255, 0), 4 );
        line( img_matches, scene_corners[1] + Point2f( img_object.cols, 0), scene_corners[2] + Point2f( img_object.cols, 0), Scalar( 0, 255, 0), 4 );
        line( img_matches, scene_corners[2] + Point2f( img_object.cols, 0), scene_corners[3] + Point2f( img_object.cols, 0), Scalar( 0, 255, 0), 4 );
        line( img_matches, scene_corners[3] + Point2f( img_object.cols, 0), scene_corners[0] + Point2f( img_object.cols, 0), Scalar( 0, 255, 0), 4 );

        //-- Show detected matches
            imshow( "AR image", dst );
            imshow( "Good Matches & Object detection", img_matches );
        }        //-- Get the corners from the image_1 ( the object to be "detected" )
    waitKey(0);    
    return 0;
}

